--  2006 Peter Thiemann
{- | Simple seat reservation using transactional variables.
-}
module Main where

import WASH.CGI.CGI hiding (map)
import WASH.CGI.CGIOutput (itell)
import qualified WASH.DB.Transaction as T

import qualified WASH.Utility.Auxiliary as Aux (appendDebugFile)
import qualified WASH.CGI.BaseCombinators as BC

import qualified Database.HDBC.PostgreSQL as DB (connectPostgreSQL)

import IO
import Maybe
import Monad


-- |debugging aid
deb str = 
  BC.unsafe_io (Aux.appendDebugFile "/tmp/DEBUG" ('\n':str))

type SeatCategory = Char
seatCategories :: [SeatCategory]
seatCategories = ['A'..'J']
seatNumbers :: [Int]
seatNumbers = [0..9]

appTitle = "WASH my Seat[DB]"

postgresConnectionString = 
  "host=localhost dbname=wash user=thiemann password=peter"

registerDB :: T.TCGI b T.D
registerDB =
  T.register postgresConnectionString DB.connectPostgreSQL
  

main = do
  catch (run mainCGI) exceptionHandler

exceptionHandler e =
  itell stdout $
  standardPage "Exception" (text ("Exception: " ++ show e))
    
mainCGI =
  deb ("> main (Reservation3)") >>
  standardQuery appTitle
  <#>
    <h3>The WASH-based reservation system</h3>
    <p>Choose a seat category:</p>
    <table border="1">
      <tr><td align="center">STAGE</td></tr>
      <% mapM_ makeCategoryButton seatCategories %> 
    </table>
  </#>

-- | create a button for seat category in overview
makeCategoryButton seatCategory =
  let carving = attr "value" ("Category " ++ [seatCategory]) in
  <tr><td><input type="submit" WASH:callback="getSeatStates seatCategory" 
		 <%carving%> /></td></tr>

-- | step 0: get current seat situation
getSeatStates :: SeatCategory -> CGI ()
getSeatStates seatCategory = do
  T.with [] (attemptReservation seatCategory) $
    \ ta ->
	do washDB <- registerDB
	   seatHandles <- initSeats washDB seatCategory
	   seatStates  <- mapM seatStatus seatHandles
	   T.setCommit ta seatStates

-- | step 1: choose from available seats
attemptReservation seatCategory seatStates =
  T.with (Left (seatCategory, seatStates)) afterReservation
    (\ ta ->
     standardReservation seatCategory seatStates ta
    )

afterReservation tstate =
  case tstate of
    Left (seatCategory, seatStates) ->
      -- failed or aborted
      commitFailed (seatCategory, seatStates)
    Right n ->
      -- reserved n seats
      goodBye n

commitFailed (seatCategory, seatStates) =
  standardQuery (appTitle ++ ": Seat Reservation Impossible")
    <#>
      <h3>We were unable to complete your seat reservation request</h3>
      <h3>Do you want to try again?</h3>
      <input type="submit" value="TRY AGAIN" 
	     WASH:callback="attemptReservation seatCategory seatStates" />
      <input type="submit" value="EXIT" WASH:callback="goodBye 0"/>
    </#>

standardReservation seatCategory seatStates ta =
  standardQuery (appTitle ++ ": Seat Offerings in Category " ++ [seatCategory])
    <#>
    <h3>Available seats</h3>
    <table>
      <tr><% buttons <- mapM makeSeatButton seatStates %></tr>
    </table>
    <% submit (FL buttons) 
   	   (chosenSeats seatCategory seatStates ta)
   	   (attr "value" "CHOOSE") %>
    </#>

makeSeatButton (seatName, seatState) =
  let myColor | seatState = "background-color: red"
	      | otherwise = "background-color: green"
      disabled | seatState = attr "disabled" "disabled"
	       | otherwise = empty
  in
  td $ do
    attr "style" myColor
    label $ do
      attr "for" seatName
      text "Seat "
      text seatName
      text " "
    checkboxInputField (disabled ## attr "id" seatName)

initSeats :: T.D -> SeatCategory -> T.TCGI b [T.T]
initSeats washDB seatCategory =
  T.findByNV washDB "seat" [("category", Just [seatCategory])] " ORDER BY name"

seatStatus seatHandle =
  do [m_name, m_cid] <- T.getMany seatHandle ["name", "customer_id"]
     return (fromJust m_name, isJust m_cid)

checkSeat seatHandle =
  do m_cid <- T.get seatHandle "customer_id"
     return (isJust m_cid)

reserveSeat :: T.D -> T.T -> Bool -> (String, x) -> T.TCGI b (Bool, Maybe T.T)
reserveSeat washDB customer occupy (seatName, _) =
  if occupy then
     do available <- T.findByNV washDB "seat"
	              [("name", Just seatName), ("customer_id", Nothing)] ""
	case available of 
	  [seatHandle] -> do
	    T.set seatHandle "customer_id" (T.getPrimaryKey customer)
	    return (True, Just seatHandle)
	  _ ->
	    return (False, Nothing)
  else
    return (True, Nothing)

updateSeatState ss (True, _) =
  ss
updateSeatState (seatName, seatState) (False, _) =
  (seatName, not seatState)

-- | step 2: submit payment information
chosenSeats seatCategory seatStates ta (FL buttons) =
  let reservations = Prelude.map value buttons in do
  washDB <- registerDB
  customer <- T.create washDB "customer" 
    [("name", Nothing), ("address", Nothing), ("payment", Nothing)]
  reservationOK <- zipWithM (reserveSeat washDB customer) reservations seatStates
  if and (map fst reservationOK) then
    standardQuery (appTitle ++ ": Enter Payment Data")
    <#>
    <% displayReservations seatCategory reservations %>
    <h3>Enter Payment Data</h3>
    <table>
      <tr><td>Name</td> <td><input type="text" name="name" /></td></tr>
      <tr><td>Address</td> <td><input type="text" name="address" /></td></tr>
      <tr><td>Acct. Info</td> <td><input type="text" name="acctinfo" /></td></tr>
    </table>
    <p><input type="submit" value="ENTER" 
	      WASH:callback="confirmReservation washDB seatCategory reservations customer ta" 
	      WASH:parms="name,address,acctinfo" />
    </p>
    </#>
   else
    let newSeatStates = zipWith updateSeatState seatStates reservationOK in
    standardQuery (appTitle ++ ": Try Again")
    <#>
      Your seats have already been taken in the meantime. Please try again.
      <br />
      <input type="submit" value="CONTINUE"
	     WASH:callback="T.abandon ta (Left (seatCategory, newSeatStates))" />
    </#>

displayReservedSeat reservation seatNumber =
  if reservation
  then text (' ':show seatNumber)
  else empty

displayReservations seatCategory reservations =
  <#>
    <h3>Your Chosen Seats</h3>
    Category <%= [seatCategory] %>,
    Seats<% zipWithM_ displayReservedSeat reservations seatNumbers %>
  </#>

-- | step 3: confirm reservation
confirmReservation washDB seatCategory reservations customer ta
  (nameU, addressU, acctinfoU) =
  let name = unNonEmpty nameU
      address = unNonEmpty addressU
      acctinfo = unNonEmpty acctinfoU
  in do
  T.setMany customer  [("name", Just name)
		      ,("address", Just address)
		      ,("payment", Just acctinfo)]
  standardQuery (appTitle ++ ": Confirm Your Data")
    <#>
    <% displayReservations seatCategory reservations %>
    <h3>Your Payment Data</h3>
    <table>
      <tr><td>Name</td> <td><%= name %></td></tr>
      <tr><td>Address</td> <td><%= address %></td></tr>
      <tr><td>Acct. Info</td> <td><%= acctinfo %></td></tr>
    </table>
    <h3>Press ENTER to Confirm Your Reservation</h3>
    <p><input type="submit" value="ENTER" 
	      WASH:callback="confirmedReservation reservations ta" />
    </p>
    </#>

-- | step 4: process confirmed reservation
confirmedReservation reservations ta =
  let nrOfSeats = sum (Prelude.map fromEnum reservations) in
  T.setCommit ta (Right nrOfSeats)

goodBye r =
  standardQuery (appTitle ++ ": Good Bye")
  <#>
  <h3>Thank you for your reservation</h3>
  You reserved <%= r %> seats with us.
  </#>

